@@ -16,7 +16,10 @@ import json
 import email
 import gzip
 
-from datetime import timedelta
+import urllib.parse
+import requests
+
+from datetime import datetime, timedelta
 
 from email import utils, charset
 from email.mime.text import MIMEText
@@ -323,19 +326,24 @@ def explode(gitdir, lmsg, mailfrom=None, retrieve_links=True, fpopts=None):
             # Is this the cover letter?
             if msubj.counter == 0:
                 # We rebuild the message from scratch
-                cmsg = MIMEMultipart()
-                cmsg.add_header('From', mailfrom)
-                cmsg.add_header('Subject', '[' + ' '.join(msubj.prefixes) + '] ' + lmsg.subject)
-                cmsg.add_header('Date', lmsg.msg.get('Date'))
-
                 # The cover letter body is the pull request body, plus a few trailers
                 body = '%s\n\nbase-commit: %s\nPR-Link: %s\n' % (
                     lmsg.body.strip(), lmsg.pr_base_commit, config['linkmask'] % lmsg.msgid)
-                cmsg.attach(MIMEText(body, 'plain'))
 
-                # now we attach the original request
-                # XXX: seems redundant, so turned off for now
-                # cmsg.attach(MIMEMessage(lmsg.msg))
+                # Make it a multipart if we're doing retrieve_links
+                if retrieve_links:
+                    cmsg = MIMEMultipart()
+                    cmsg.attach(MIMEText(body, 'plain'))
+                else:
+                    cmsg = email.message.EmailMessage()
+                    cmsg.set_payload(body)
+
+                cmsg.add_header('From', mailfrom)
+                cmsg.add_header('Subject', '[' + ' '.join(msubj.prefixes) + '] ' + lmsg.subject)
+                cmsg.add_header('Date', lmsg.msg.get('Date'))
+                cmsg.set_charset('utf-8')
+                cmsg.replace_header('Content-Transfer-Encoding', '8bit')
+
                 msg = cmsg
 
             else:
@@ -427,6 +435,72 @@ def explode(gitdir, lmsg, mailfrom=None, retrieve_links=True, fpopts=None):
     return msgs
 
 
+def get_pr_from_github(ghurl: str):
+    loc = urllib.parse.urlparse(ghurl)
+    chunks = loc.path.strip('/').split('/')
+    rproj = chunks[0]
+    rrepo = chunks[1]
+    rpull = chunks[-1]
+    apiurl = f'https://api.github.com/repos/{rproj}/{rrepo}/pulls/{rpull}'
+    req = requests.session()
+    # Do we have a github API key?
+    config = b4.get_main_config()
+    ghkey = config.get('gh-api-key')
+    if ghkey:
+        req.headers.update({'Authorization': f'token {ghkey}'})
+    req.headers.update({'Accept': 'application/vnd.github.v3+json'})
+    resp = req.get(apiurl)
+    if resp.status_code != 200:
+        logger.critical('Server returned an error: %s', resp.status_code)
+        return None
+    prdata = resp.json()
+
+    msg = email.message.EmailMessage()
+    msg.set_payload(prdata.get('body', '(no body)'))
+    head = prdata.get('head', {})
+    repo = head.get('repo', {})
+    base = prdata.get('base', {})
+    user = prdata.get('user', {})
+
+    ulogin = user.get('login')
+    fake_email = f'{ulogin}@github.com'
+    apiurl = f'https://api.github.com/users/{ulogin}'
+    resp = req.get(apiurl)
+    if resp.status_code == 200:
+        udata = resp.json()
+        uname = udata.get('name')
+        if not uname:
+            uname = ulogin
+        uemail = udata.get('email')
+        if not uemail:
+            uemail = fake_email
+    else:
+        uname = ulogin
+        uemail = fake_email
+
+    msg['From'] = f'{uname} <{uemail}>'
+    title = prdata.get('title', '')
+    msg['Subject'] = f'[GIT PULL] {title}'
+    msg['To'] = fake_email
+    msg['Message-Id'] = utils.make_msgid(idstring=f'{rproj}-{rrepo}-pr-{rpull}', domain='github.com')
+    created_at = utils.format_datetime(datetime.strptime(prdata.get('created_at'), '%Y-%m-%dT%H:%M:%SZ'))
+    msg['Date'] = created_at
+    # We are going to turn it into bytes and then parse again
+    # in order to avoid bugs with python's message parsing routines that
+    # end up not doing the right thing when decoding 8bit message bodies
+    msg.set_charset('utf-8')
+    msg.replace_header('Content-Transfer-Encoding', '8bit')
+    bug_avoidance = msg.as_string(policy=b4.emlpolicy).encode()
+    cmsg = email.message_from_bytes(bug_avoidance)
+    lmsg = b4.LoreMessage(cmsg)
+    lmsg.pr_base_commit = base.get('sha')
+    lmsg.pr_repo = repo.get('clone_url')
+    lmsg.pr_ref = head.get('ref')
+    lmsg.pr_tip_commit = head.get('sha')
+    lmsg.pr_remote_tip_commit = head.get('sha')
+    return lmsg
+
+
 def main(cmdargs):
     gitdir = cmdargs.gitdir
     lmsg = None
@@ -434,36 +508,33 @@ def main(cmdargs):
     if not sys.stdin.isatty():
         logger.debug('Getting PR message from stdin')
         msg = email.message_from_bytes(sys.stdin.buffer.read())
-        msgid = b4.LoreMessage.get_clean_msgid(msg)
+        cmdargs.msgid = b4.LoreMessage.get_clean_msgid(msg)
         lmsg = parse_pr_data(msg)
     else:
-        logger.debug('Getting PR message from public-inbox')
+        if cmdargs.msgid and 'github.com' in cmdargs.msgid and '/pull/' in cmdargs.msgid:
+            logger.debug('Getting PR info from Github')
+            lmsg = get_pr_from_github(cmdargs.msgid)
+        else:
+            logger.debug('Getting PR message from public-inbox')
 
-        msgid = b4.get_msgid(cmdargs)
-        msgs = b4.get_pi_thread_by_msgid(msgid)
-        if not msgs:
-            return
-        for msg in msgs:
-            mmsgid = b4.LoreMessage.get_clean_msgid(msg)
-            if mmsgid == msgid:
-                lmsg = parse_pr_data(msg)
-                break
+            msgid = b4.get_msgid(cmdargs)
+            msgs = b4.get_pi_thread_by_msgid(msgid)
+            if not msgs:
+                return
+            for msg in msgs:
+                mmsgid = b4.LoreMessage.get_clean_msgid(msg)
+                if mmsgid == msgid:
+                    lmsg = parse_pr_data(msg)
+                    break
 
     if lmsg is None or lmsg.pr_remote_tip_commit is None:
-        logger.critical('ERROR: Could not find pull request info in %s', msgid)
+        logger.critical('ERROR: Could not find pull request info in %s', cmdargs.msgid)
         sys.exit(1)
 
     if not lmsg.pr_tip_commit:
         lmsg.pr_tip_commit = lmsg.pr_remote_tip_commit
 
     if cmdargs.explode:
-        savefile = cmdargs.outmbox
-        if savefile is None:
-            savefile = '%s.mbx' % lmsg.msgid
-        if os.path.exists(savefile):
-            logger.info('File exists: %s', savefile)
-            sys.exit(1)
-
         # Set up a temporary clone
         with b4.git_temp_clone(gitdir) as tc:
             try:
@@ -472,18 +543,58 @@ def main(cmdargs):
                 logger.critical('Nothing exploded.')
                 sys.exit(1)
 
-            if msgs:
-                smbx = mailbox.mbox(savefile)
-                for msg in msgs:
-                    smbx.add(msg.as_string(policy=b4.emlpolicy).encode())
-                smbx.close()
-                logger.info('---')
-                logger.info('Saved %s', savefile)
-                sys.exit(0)
+        if msgs:
+            if cmdargs.sendidentity:
+                # Pass exploded series via git-send-email
+                config = b4.get_config_from_git(rf'sendemail\.{cmdargs.sendidentity}\..*')
+                if not len(config):
+                    logger.critical('Not able to find sendemail.%s configuration', cmdargs.sendidentity)
+                    sys.exit(1)
+                # Make sure from is not overridden by current user
+                mailfrom = msgs[0].get('from')
+                gitargs = ['send-email', '--identity', cmdargs.sendidentity, '--from', mailfrom]
+                if cmdargs.dryrun:
+                    gitargs.append('--dry-run')
+                # Write out everything into a temporary dir
+                counter = 0
+                with tempfile.TemporaryDirectory() as tfd:
+                    for msg in msgs:
+                        outfile = os.path.join(tfd, '%04d' % counter)
+                        with open(outfile, 'wb') as tfh:
+                            tfh.write(msg.as_string(policy=b4.emlpolicy).encode())
+                        gitargs.append(outfile)
+                        counter += 1
+                    ecode, out = b4.git_run_command(cmdargs.gitdir, gitargs, logstderr=True)
+                    if cmdargs.dryrun:
+                        logger.info(out)
+                    sys.exit(ecode)
+
+            config = b4.get_main_config()
+            if config.get('save-maildirs', 'no') == 'yes':
+                save_maildir = True
+                dftext = 'maildir'
             else:
-                logger.critical('Nothing exploded.')
+                save_maildir = False
+                dftext = 'mbx'
+            savefile = cmdargs.outmbox
+            if savefile is None:
+                savefile = f'{lmsg.msgid}.{dftext}'
+            if os.path.exists(savefile):
+                logger.info('File exists: %s', savefile)
                 sys.exit(1)
 
+            if save_maildir:
+                b4.save_maildir(msgs, savefile)
+            else:
+                with open(savefile, 'w') as fh:
+                    b4.save_git_am_mbox(msgs, fh)
+            logger.info('---')
+            logger.info('Saved %s', savefile)
+            sys.exit(0)
+        else:
+            logger.critical('Nothing exploded.')
+            sys.exit(1)
+
     exists = b4.git_commit_exists(gitdir, lmsg.pr_tip_commit)
     if exists:
         # Is it in any branch, or just flapping in the wind?
